@@ -84,22 +84,34 @@
 //! Fields can also be annotated with `#[default(VALUE)]` to indicate ASN.1
 //! `OPTIONAL DEFAULT` values. In this case, the field's type should be `T`,
 //! and not `Option<T>`.
+//!
+//! These derives may also be used with `enum`s to generate `CHOICE`
+//! implementations.
+//! ```text
+//! #[derive(asn1::Asn1Read, asn1::Asn1Write)]
+//! enum Time {
+//!     UTCTime(asn1::UtcTime),
+//!     GeneralizedTime(asn1::GeneralizedTime)
+//! }
+//! ```
+//!
+//! All variants must have a single un-named field.
 
 extern crate alloc;
 
 mod bit_string;
-mod object_identitifer;
+mod object_identifier;
 mod parser;
 mod types;
 mod writer;
 
 pub use crate::bit_string::BitString;
-pub use crate::object_identitifer::ObjectIdentifier;
+pub use crate::object_identifier::ObjectIdentifier;
 pub use crate::parser::{parse, parse_single, ParseError, ParseResult, Parser};
 pub use crate::types::{
-    Asn1Readable, Asn1Writable, BigUint, Choice1, Choice2, Choice3, Enumerated, IA5String,
-    PrintableString, Sequence, SequenceOf, SequenceOfWriter, SequenceWriter, SetOf, SetOfWriter,
-    SimpleAsn1Readable, SimpleAsn1Writable, Tlv, UtcTime,
+    Asn1Readable, Asn1Writable, BigUint, Choice1, Choice2, Choice3, Enumerated, GeneralizedTime,
+    IA5String, PrintableString, Sequence, SequenceOf, SequenceOfWriter, SequenceWriter, SetOf,
+    SetOfWriter, SimpleAsn1Readable, SimpleAsn1Writable, Tlv, UtcTime, Utf8String, VisibleString,
 };
 #[cfg(feature = "const-generics")]
 pub use crate::types::{Explicit, Implicit};
@@ -127,3 +139,17 @@ pub fn to_optional_default<'a, T: PartialEq>(v: &'a T, default: &'a T) -> Option
         Some(v)
     }
 }
+
+/// This API is public so that it may be used from macros, but should not be
+/// considered a part of the supported API surface.
+#[doc(hidden)]
+pub const fn implicit_tag(tag: u8, inner_tag: u8) -> u8 {
+    types::CONTEXT_SPECIFIC | tag | (inner_tag & types::CONSTRUCTED)
+}
+
+/// This API is public so that it may be used from macros, but should not be
+/// considered a part of the supported API surface.
+#[doc(hidden)]
+pub const fn explicit_tag(tag: u8) -> u8 {
+    types::CONTEXT_SPECIFIC | types::CONSTRUCTED | tag
+}
