@@ -1,4 +1,4 @@
-use crate::types::{Asn1Readable, SimpleAsn1Readable, Tlv, CONSTRUCTED, CONTEXT_SPECIFIC};
+use crate::types::{Asn1Readable, SimpleAsn1Readable, Tlv};
 
 /// ParseError are returned when there is an error parsing the ASN.1 data.
 #[derive(Debug, PartialEq)]
@@ -155,7 +155,7 @@ impl<'a> Parser<'a> {
         &mut self,
         tag: u8,
     ) -> ParseResult<Option<T>> {
-        let expected_tag = CONTEXT_SPECIFIC | CONSTRUCTED | tag;
+        let expected_tag = crate::explicit_tag(tag);
         if self.peek_u8() != Some(expected_tag) {
             return Ok(None);
         }
@@ -169,7 +169,7 @@ impl<'a> Parser<'a> {
         &mut self,
         tag: u8,
     ) -> ParseResult<Option<T>> {
-        let expected_tag = CONTEXT_SPECIFIC | tag | (T::TAG & CONSTRUCTED);
+        let expected_tag = crate::implicit_tag(tag, T::TAG);
         if self.peek_u8() != Some(expected_tag) {
             return Ok(None);
         }
@@ -183,8 +183,9 @@ mod tests {
     use super::Parser;
     use crate::types::Asn1Readable;
     use crate::{
-        BigUint, BitString, Choice1, Choice2, Choice3, Enumerated, IA5String, ObjectIdentifier,
-        ParseError, ParseResult, PrintableString, Sequence, SequenceOf, SetOf, Tlv, UtcTime,
+        BigUint, BitString, Choice1, Choice2, Choice3, Enumerated, GeneralizedTime, IA5String,
+        ObjectIdentifier, ParseError, ParseResult, PrintableString, Sequence, SequenceOf, SetOf,
+        Tlv, UtcTime, Utf8String, VisibleString,
     };
     #[cfg(feature = "const-generics")]
     use crate::{Explicit, Implicit};
@@ -428,7 +429,7 @@ mod tests {
     }
 
     #[test]
-    fn test_parse_object_identitifer() {
+    fn test_parse_object_identifier() {
         assert_parses::<ObjectIdentifier<'_>>(&[
             (
                 Ok(ObjectIdentifier::from_string("2.5").unwrap()),
@@ -498,6 +499,24 @@ mod tests {
         ])
     }
 
+    #[test]
+    fn test_parse_utf8string() {
+        assert_parses::<Utf8String>(&[
+            (Ok(Utf8String::new("abc")), b"\x0c\x03abc"),
+            (Ok(Utf8String::new(")")), b"\x0c\x01)"),
+            (Err(ParseError::InvalidValue), b"\x0c\x01\xff"),
+        ])
+    }
+
+    #[test]
+    fn test_parse_visiblestring() {
+        assert_parses::<VisibleString>(&[
+            (Ok(VisibleString::new("abc").unwrap()), b"\x1a\x03abc"),
+            (Ok(VisibleString::new(")").unwrap()), b"\x1a\x01)"),
+            (Err(ParseError::InvalidValue), b"\x1a\x01\n"),
+        ])
+    }
+
     #[test]
     fn test_parse_utctime() {
         assert_parses::<UtcTime>(&[
@@ -560,6 +579,62 @@ mod tests {
         ]);
     }
 
+    #[test]
+    fn test_generalizedtime() {
+        assert_parses::<GeneralizedTime>(&[
+            (
+                Ok(GeneralizedTime::new(Utc.ymd(2010, 1, 2).and_hms(3, 4, 5))),
+                b"\x18\x0f20100102030405Z",
+            ),
+            (
+                Ok(GeneralizedTime::new(
+                    FixedOffset::east(6 * 60 * 60 + 7 * 60)
+                        .ymd(2010, 1, 2)
+                        .and_hms(3, 4, 5)
+                        .into(),
+                )),
+                b"\x18\x1320100102030405+0607",
+            ),
+            (
+                Ok(GeneralizedTime::new(
+                    FixedOffset::west(6 * 60 * 60 + 7 * 60)
+                        .ymd(2010, 1, 2)
+                        .and_hms(3, 4, 5)
+                        .into(),
+                )),
+                b"\x18\x1320100102030405-0607",
+            ),
+            (Err(ParseError::InvalidValue), b"\x18\x0e20100102030405"),
+            (Err(ParseError::InvalidValue), b"\x18\x0e00000100000000Z"),
+            (Err(ParseError::InvalidValue), b"\x18\x0e20101302030405Z"),
+            (Err(ParseError::InvalidValue), b"\x18\x0e20100002030405Z"),
+            (Err(ParseError::InvalidValue), b"\x18\x0e20100100030405Z"),
+            (Err(ParseError::InvalidValue), b"\x18\x0e20100132030405Z"),
+            (Err(ParseError::InvalidValue), b"\x18\x0e20100231030405Z"),
+            (Err(ParseError::InvalidValue), b"\x18\x0e20100102240405Z"),
+            (Err(ParseError::InvalidValue), b"\x18\x0e20100102036005Z"),
+            (Err(ParseError::InvalidValue), b"\x18\x0e20100102030460Z"),
+            (Err(ParseError::InvalidValue), b"\x18\x0f-20100102030410Z"),
+            (Err(ParseError::InvalidValue), b"\x18\x0f2010-0102030410Z"),
+            (Err(ParseError::InvalidValue), b"\x18\x0f2010-0002030410Z"),
+            (Err(ParseError::InvalidValue), b"\x18\x0f201001-02030410Z"),
+            (Err(ParseError::InvalidValue), b"\x18\x0f20100102-030410Z"),
+            (Err(ParseError::InvalidValue), b"\x18\x0f2010010203-0410Z"),
+            (Err(ParseError::InvalidValue), b"\x18\x0f201001020304-10Z"),
+            // Tests for fractional seconds, which we currently don't support
+            (
+                Err(ParseError::InvalidValue),
+                b"\x18\x1620100102030405.123456Z",
+            ),
+            (
+                Err(ParseError::InvalidValue),
+                b"\x18\x1520100102030405.123456",
+            ),
+            (Err(ParseError::InvalidValue), b"\x18\x1020100102030405.Z"),
+            (Err(ParseError::InvalidValue), b"\x18\x0f20100102030405."),
+        ]);
+    }
+
     #[test]
     fn test_enumerated() {
         assert_parses::<Enumerated>(&[
@@ -709,7 +784,7 @@ mod tests {
     fn test_choice1() {
         assert_parses::<Choice1<bool>>(&[
             (Ok(Choice1::ChoiceA(true)), b"\x01\x01\xff"),
-            (Err(ParseError::UnexpectedTag { actual: 0x03 }), b"\x03"),
+            (Err(ParseError::UnexpectedTag { actual: 0x03 }), b"\x03\x00"),
             (Err(ParseError::ShortData), b""),
         ]);
     }
@@ -719,7 +794,7 @@ mod tests {
         assert_parses::<Choice2<bool, i64>>(&[
             (Ok(Choice2::ChoiceA(true)), b"\x01\x01\xff"),
             (Ok(Choice2::ChoiceB(18)), b"\x02\x01\x12"),
-            (Err(ParseError::UnexpectedTag { actual: 0x03 }), b"\x03"),
+            (Err(ParseError::UnexpectedTag { actual: 0x03 }), b"\x03\x00"),
             (Err(ParseError::ShortData), b""),
         ]);
     }
@@ -730,7 +805,7 @@ mod tests {
             (Ok(Choice3::ChoiceA(true)), b"\x01\x01\xff"),
             (Ok(Choice3::ChoiceB(18)), b"\x02\x01\x12"),
             (Ok(Choice3::ChoiceC(())), b"\x05\x00"),
-            (Err(ParseError::UnexpectedTag { actual: 0x03 }), b"\x03"),
+            (Err(ParseError::UnexpectedTag { actual: 0x03 }), b"\x03\x00"),
             (Err(ParseError::ShortData), b""),
         ]);
     }
