@@ -285,6 +285,92 @@ impl<'a> SimpleAsn1Writable<'a> for IA5String<'a> {
     }
 }
 
+/// Type for use with `Parser.read_element` and `Writer.write_element` for
+/// handling ASN.1 `UTF8String`.
+#[derive(Clone, Debug, PartialEq)]
+pub struct Utf8String<'a>(&'a str);
+
+impl<'a> Utf8String<'a> {
+    pub fn new(s: &'a str) -> Utf8String<'a> {
+        Utf8String(s)
+    }
+
+    fn new_from_bytes(s: &'a [u8]) -> Option<Utf8String> {
+        Some(Utf8String(core::str::from_utf8(s).ok()?))
+    }
+
+    pub fn as_str(&self) -> &'a str {
+        self.0
+    }
+}
+
+impl<'a> SimpleAsn1Readable<'a> for Utf8String<'a> {
+    const TAG: u8 = 0x0c;
+    fn parse_data(data: &'a [u8]) -> ParseResult<Self> {
+        Utf8String::new_from_bytes(data).ok_or(ParseError::InvalidValue)
+    }
+}
+impl<'a> SimpleAsn1Writable<'a> for Utf8String<'a> {
+    const TAG: u8 = 0x0c;
+    fn write_data(&self, dest: &mut Vec<u8>) {
+        dest.extend_from_slice(self.0.as_bytes());
+    }
+}
+
+/// Type for use with `Parser.read_element` and `Writer.write_element` for
+/// handling ASN.1 `VisibleString`.  An `VisibleString` contains an `&str`
+/// with only valid characers.
+#[derive(Clone, Debug, PartialEq)]
+pub struct VisibleString<'a>(&'a str);
+
+impl<'a> VisibleString<'a> {
+    pub fn new(s: &'a str) -> Option<VisibleString<'a>> {
+        if VisibleString::verify(s.as_bytes()) {
+            Some(VisibleString(s))
+        } else {
+            None
+        }
+    }
+
+    fn new_from_bytes(s: &'a [u8]) -> Option<VisibleString> {
+        if VisibleString::verify(s) {
+            // TODO: This value is always valid utf-8 because we just verified
+            // the contents, but I don't want to call an unsafe function, so we
+            // end up validating it twice. If your profile says this is slow,
+            // now you know why.
+            Some(VisibleString(core::str::from_utf8(s).unwrap()))
+        } else {
+            None
+        }
+    }
+
+    fn verify(s: &[u8]) -> bool {
+        for b in s {
+            if !(b.is_ascii_graphic() || *b == b' ') {
+                return false;
+            }
+        }
+        true
+    }
+
+    pub fn as_str(&self) -> &'a str {
+        self.0
+    }
+}
+
+impl<'a> SimpleAsn1Readable<'a> for VisibleString<'a> {
+    const TAG: u8 = 0x1a;
+    fn parse_data(data: &'a [u8]) -> ParseResult<Self> {
+        VisibleString::new_from_bytes(data).ok_or(ParseError::InvalidValue)
+    }
+}
+impl<'a> SimpleAsn1Writable<'a> for VisibleString<'a> {
+    const TAG: u8 = 0x1a;
+    fn write_data(&self, dest: &mut Vec<u8>) {
+        dest.extend_from_slice(self.0.as_bytes());
+    }
+}
+
 fn validate_integer(data: &[u8], signed: bool) -> ParseResult<()> {
     if data.is_empty() {
         return Err(ParseError::InvalidValue);
@@ -493,6 +579,13 @@ fn push_two_digits(dest: &mut Vec<u8>, val: u8) {
     dest.push(b'0' + (val % 10));
 }
 
+fn push_four_digits(dest: &mut Vec<u8>, val: u16) {
+    dest.push(b'0' + ((val / 1000) % 10) as u8);
+    dest.push(b'0' + ((val / 100) % 10) as u8);
+    dest.push(b'0' + ((val / 10) % 10) as u8);
+    dest.push(b'0' + (val % 10) as u8);
+}
+
 impl SimpleAsn1Writable<'_> for UtcTime {
     const TAG: u8 = 0x17;
     fn write_data(&self, dest: &mut Vec<u8>) {
@@ -515,6 +608,51 @@ impl SimpleAsn1Writable<'_> for UtcTime {
     }
 }
 
+/// Used for parsing and writing ASN.1 `GENERALIZED TIME` values. Wraps a
+/// `chrono::DateTime<Utc>`.
+#[derive(Debug, PartialEq)]
+pub struct GeneralizedTime(chrono::DateTime<chrono::Utc>);
+
+impl GeneralizedTime {
+    pub fn new(v: chrono::DateTime<chrono::Utc>) -> GeneralizedTime {
+        GeneralizedTime(v)
+    }
+
+    pub fn as_chrono(&self) -> &chrono::DateTime<chrono::Utc> {
+        &self.0
+    }
+}
+
+impl SimpleAsn1Readable<'_> for GeneralizedTime {
+    const TAG: u8 = 0x18;
+    fn parse_data(data: &[u8]) -> ParseResult<GeneralizedTime> {
+        let data = core::str::from_utf8(data).map_err(|_| ParseError::InvalidValue)?;
+        if let Ok(v) = chrono::Utc.datetime_from_str(data, "%Y%m%d%H%M%SZ") {
+            return Ok(GeneralizedTime::new(v));
+        }
+        if let Ok(v) = chrono::DateTime::parse_from_str(data, "%Y%m%d%H%M%S%z") {
+            return Ok(GeneralizedTime::new(v.into()));
+        }
+
+        Err(ParseError::InvalidValue)
+    }
+}
+
+impl SimpleAsn1Writable<'_> for GeneralizedTime {
+    const TAG: u8 = 0x18;
+    fn write_data(&self, dest: &mut Vec<u8>) {
+        push_four_digits(dest, self.0.year().try_into().unwrap());
+        push_two_digits(dest, self.0.month().try_into().unwrap());
+        push_two_digits(dest, self.0.day().try_into().unwrap());
+
+        push_two_digits(dest, self.0.hour().try_into().unwrap());
+        push_two_digits(dest, self.0.minute().try_into().unwrap());
+        push_two_digits(dest, self.0.second().try_into().unwrap());
+
+        dest.push(b'Z');
+    }
+}
+
 /// An ASN.1 `ENUMERATED` value.
 #[derive(Debug, PartialEq)]
 pub struct Enumerated(u32);
@@ -588,14 +726,13 @@ macro_rules! declare_choice {
             )*
         > Asn1Readable<'a> for $count<$($number,)*> {
             fn parse(parser: &mut Parser<'a>) -> ParseResult<Self> {
-                let tag = parser.peek_u8();
-                match tag {
-                    $(
-                        Some(tag) if $number::can_parse(tag) => Ok($count::$name(parser.read_element::<$number>()?)),
-                    )*
-                    Some(tag) => Err(ParseError::UnexpectedTag{actual: tag}),
-                    None => Err(ParseError::ShortData),
-                }
+                let tlv = parser.read_tlv()?;
+                $(
+                    if $number::can_parse(tlv.tag()) {
+                        return Ok($count::$name(tlv.parse::<$number>()?));
+                    }
+                )*
+                Err(ParseError::UnexpectedTag{actual: tlv.tag()})
             }
 
             fn can_parse(tag: u8) -> bool {
@@ -611,7 +748,7 @@ macro_rules! declare_choice {
         impl<
             'a,
             $(
-                $number: SimpleAsn1Writable<'a>,
+                $number: Asn1Writable<'a>,
             )*
         > Asn1Writable<'a> for $count<$($number,)*> {
             fn write(&self, w: &mut Writer) {
@@ -691,12 +828,12 @@ impl<'a> SimpleAsn1Writable<'a> for SequenceWriter<'a> {
 
 /// Represents an ASN.1 `SEQUENCE OF`. This is an `Iterator` over values that
 /// are decoded.
-pub struct SequenceOf<'a, T: SimpleAsn1Readable<'a>> {
+pub struct SequenceOf<'a, T: Asn1Readable<'a>> {
     parser: Parser<'a>,
     _phantom: PhantomData<T>,
 }
 
-impl<'a, T: SimpleAsn1Readable<'a>> SequenceOf<'a, T> {
+impl<'a, T: Asn1Readable<'a>> SequenceOf<'a, T> {
     #[inline]
     pub(crate) fn new(data: &'a [u8]) -> SequenceOf<'a, T> {
         SequenceOf {
@@ -706,7 +843,7 @@ impl<'a, T: SimpleAsn1Readable<'a>> SequenceOf<'a, T> {
     }
 }
 
-impl<'a, T: SimpleAsn1Readable<'a>> Clone for SequenceOf<'a, T> {
+impl<'a, T: Asn1Readable<'a>> Clone for SequenceOf<'a, T> {
     fn clone(&self) -> SequenceOf<'a, T> {
         SequenceOf {
             parser: self.parser.clone_internal(),
@@ -715,7 +852,7 @@ impl<'a, T: SimpleAsn1Readable<'a>> Clone for SequenceOf<'a, T> {
     }
 }
 
-impl<'a, T: SimpleAsn1Readable<'a> + 'a> SimpleAsn1Readable<'a> for SequenceOf<'a, T> {
+impl<'a, T: Asn1Readable<'a> + 'a> SimpleAsn1Readable<'a> for SequenceOf<'a, T> {
     const TAG: u8 = 0x10 | CONSTRUCTED;
     #[inline]
     fn parse_data(data: &'a [u8]) -> ParseResult<Self> {
@@ -729,7 +866,7 @@ impl<'a, T: SimpleAsn1Readable<'a> + 'a> SimpleAsn1Readable<'a> for SequenceOf<'
     }
 }
 
-impl<'a, T: SimpleAsn1Readable<'a>> Iterator for SequenceOf<'a, T> {
+impl<'a, T: Asn1Readable<'a>> Iterator for SequenceOf<'a, T> {
     type Item = T;
 
     fn next(&mut self) -> Option<Self::Item> {
@@ -744,9 +881,7 @@ impl<'a, T: SimpleAsn1Readable<'a>> Iterator for SequenceOf<'a, T> {
     }
 }
 
-impl<'a, T: SimpleAsn1Readable<'a> + SimpleAsn1Writable<'a>> SimpleAsn1Writable<'a>
-    for SequenceOf<'a, T>
-{
+impl<'a, T: Asn1Readable<'a> + Asn1Writable<'a>> SimpleAsn1Writable<'a> for SequenceOf<'a, T> {
     const TAG: u8 = 0x10 | CONSTRUCTED;
     fn write_data(&self, dest: &mut Vec<u8>) {
         let mut w = Writer::new(dest);
@@ -779,12 +914,12 @@ impl<'a, T: Asn1Writable<'a>> SimpleAsn1Writable<'a> for SequenceOfWriter<'a, T>
 
 /// Represents an ASN.1 `SET OF`. This is an `Iterator` over values that
 /// are decoded.
-pub struct SetOf<'a, T: SimpleAsn1Readable<'a>> {
+pub struct SetOf<'a, T: Asn1Readable<'a>> {
     parser: Parser<'a>,
     _phantom: PhantomData<T>,
 }
 
-impl<'a, T: SimpleAsn1Readable<'a>> SetOf<'a, T> {
+impl<'a, T: Asn1Readable<'a>> SetOf<'a, T> {
     #[inline]
     pub(crate) fn new(data: &'a [u8]) -> SetOf<'a, T> {
         SetOf {
@@ -794,7 +929,16 @@ impl<'a, T: SimpleAsn1Readable<'a>> SetOf<'a, T> {
     }
 }
 
-impl<'a, T: SimpleAsn1Readable<'a> + 'a> SimpleAsn1Readable<'a> for SetOf<'a, T> {
+impl<'a, T: Asn1Readable<'a>> Clone for SetOf<'a, T> {
+    fn clone(&self) -> SetOf<'a, T> {
+        SetOf {
+            parser: self.parser.clone_internal(),
+            _phantom: PhantomData,
+        }
+    }
+}
+
+impl<'a, T: Asn1Readable<'a> + 'a> SimpleAsn1Readable<'a> for SetOf<'a, T> {
     const TAG: u8 = 0x11 | CONSTRUCTED;
 
     #[inline]
@@ -817,7 +961,7 @@ impl<'a, T: SimpleAsn1Readable<'a> + 'a> SimpleAsn1Readable<'a> for SetOf<'a, T>
     }
 }
 
-impl<'a, T: SimpleAsn1Readable<'a>> Iterator for SetOf<'a, T> {
+impl<'a, T: Asn1Readable<'a>> Iterator for SetOf<'a, T> {
     type Item = T;
 
     fn next(&mut self) -> Option<Self::Item> {
@@ -832,6 +976,18 @@ impl<'a, T: SimpleAsn1Readable<'a>> Iterator for SetOf<'a, T> {
     }
 }
 
+impl<'a, T: Asn1Readable<'a> + Asn1Writable<'a>> SimpleAsn1Writable<'a> for SetOf<'a, T> {
+    const TAG: u8 = 0x11 | CONSTRUCTED;
+    fn write_data(&self, dest: &mut Vec<u8>) {
+        let mut w = Writer::new(dest);
+        // We are known to be ordered correctly because that's an invariant for
+        // `self`, so we don't need to sort here.
+        for el in self.clone() {
+            w.write_element(&el);
+        }
+    }
+}
+
 /// Writes an ASN.1 `SET OF` whose contents is a slice of `T`. This type is
 /// responsible for ensure the values are properly ordered when written as DER.
 pub struct SetOfWriter<'a, T: Asn1Writable<'a>> {
@@ -913,7 +1069,7 @@ impl<'a, T, const TAG: u8> From<T> for Implicit<'a, T, { TAG }> {
 impl<'a, T: SimpleAsn1Readable<'a>, const TAG: u8> SimpleAsn1Readable<'a>
     for Implicit<'a, T, { TAG }>
 {
-    const TAG: u8 = CONTEXT_SPECIFIC | TAG | (T::TAG & CONSTRUCTED);
+    const TAG: u8 = crate::implicit_tag(TAG, T::TAG);
     fn parse_data(data: &'a [u8]) -> ParseResult<Self> {
         Ok(Implicit::new(T::parse_data(data)?))
     }
@@ -923,7 +1079,7 @@ impl<'a, T: SimpleAsn1Readable<'a>, const TAG: u8> SimpleAsn1Readable<'a>
 impl<'a, T: SimpleAsn1Writable<'a>, const TAG: u8> SimpleAsn1Writable<'a>
     for Implicit<'a, T, { TAG }>
 {
-    const TAG: u8 = CONTEXT_SPECIFIC | TAG | (T::TAG & CONSTRUCTED);
+    const TAG: u8 = crate::implicit_tag(TAG, T::TAG);
 
     fn write_data(&self, dest: &mut Vec<u8>) {
         self.inner.write_data(dest);
@@ -966,7 +1122,7 @@ impl<'a, T, const TAG: u8> From<T> for Explicit<'a, T, { TAG }> {
 
 #[cfg(feature = "const-generics")]
 impl<'a, T: Asn1Readable<'a>, const TAG: u8> SimpleAsn1Readable<'a> for Explicit<'a, T, { TAG }> {
-    const TAG: u8 = CONTEXT_SPECIFIC | CONSTRUCTED | TAG;
+    const TAG: u8 = crate::explicit_tag(TAG);
     fn parse_data(data: &'a [u8]) -> ParseResult<Self> {
         Ok(Explicit::new(parse(data, |p| p.read_element::<T>())?))
     }
@@ -974,7 +1130,7 @@ impl<'a, T: Asn1Readable<'a>, const TAG: u8> SimpleAsn1Readable<'a> for Explicit
 
 #[cfg(feature = "const-generics")]
 impl<'a, T: Asn1Writable<'a>, const TAG: u8> SimpleAsn1Writable<'a> for Explicit<'a, T, { TAG }> {
-    const TAG: u8 = CONTEXT_SPECIFIC | CONSTRUCTED | TAG;
+    const TAG: u8 = crate::explicit_tag(TAG);
     fn write_data(&self, dest: &mut Vec<u8>) {
         Writer::new(dest).write_element(&self.inner);
     }
