@@ -1,4 +1,4 @@
-use crate::types::{Asn1Writable, SimpleAsn1Writable, CONSTRUCTED, CONTEXT_SPECIFIC};
+use crate::types::{Asn1Writable, SimpleAsn1Writable};
 use alloc::vec;
 use alloc::vec::Vec;
 
@@ -46,7 +46,7 @@ impl Writer<'_> {
         tag: u8,
     ) {
         if let Some(v) = val {
-            let tag = CONTEXT_SPECIFIC | CONSTRUCTED | tag;
+            let tag = crate::explicit_tag(tag);
             self.write_tlv(tag, |dest| Writer::new(dest).write_element(v));
         }
     }
@@ -59,7 +59,7 @@ impl Writer<'_> {
         tag: u8,
     ) {
         if let Some(v) = val {
-            let tag = CONTEXT_SPECIFIC | tag | (T::TAG & CONSTRUCTED);
+            let tag = crate::implicit_tag(tag, T::TAG);
             self.write_tlv(tag, |dest| v.write_data(dest));
         }
     }
@@ -115,9 +115,9 @@ mod tests {
     use super::{_insert_at_position, write, write_single, Writer};
     use crate::types::Asn1Writable;
     use crate::{
-        parse_single, BigUint, BitString, Choice1, Choice2, Choice3, Enumerated, IA5String,
-        ObjectIdentifier, PrintableString, Sequence, SequenceOf, SequenceOfWriter, SequenceWriter,
-        SetOfWriter, Tlv, UtcTime,
+        parse_single, BigUint, BitString, Choice1, Choice2, Choice3, Enumerated, GeneralizedTime,
+        IA5String, ObjectIdentifier, PrintableString, Sequence, SequenceOf, SequenceOfWriter,
+        SequenceWriter, SetOf, SetOfWriter, Tlv, UtcTime, Utf8String, VisibleString,
     };
     #[cfg(feature = "const-generics")]
     use crate::{Explicit, Implicit};
@@ -191,6 +191,34 @@ mod tests {
         ]);
     }
 
+    #[test]
+    fn test_write_utf8string() {
+        assert_writes::<Utf8String>(&[
+            (
+                Utf8String::new("Test User 1"),
+                b"\x0c\x0bTest User 1",
+            ),
+            (
+                Utf8String::new("xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"),
+                b"\x0c\x81\x80xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
+            ),
+        ]);
+    }
+
+    #[test]
+    fn test_write_visiblestring() {
+        assert_writes::<VisibleString>(&[
+            (
+                VisibleString::new("Test User 1").unwrap(),
+                b"\x1a\x0bTest User 1",
+            ),
+            (
+                VisibleString::new("xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx").unwrap(),
+                b"\x1a\x81\x80xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
+            ),
+        ]);
+    }
+
     #[test]
     fn test_write_i64() {
         assert_writes::<i64>(&[
@@ -278,6 +306,24 @@ mod tests {
         ]);
     }
 
+    #[test]
+    fn test_write_generalizedtime() {
+        assert_writes(&[
+            (
+                GeneralizedTime::new(Utc.ymd(1991, 5, 6).and_hms(23, 45, 40)),
+                b"\x18\x0f19910506234540Z",
+            ),
+            (
+                GeneralizedTime::new(Utc.timestamp(0, 0)),
+                b"\x18\x0f19700101000000Z",
+            ),
+            (
+                GeneralizedTime::new(Utc.timestamp(1258325776, 0)),
+                b"\x18\x0f20091115225616Z",
+            ),
+        ]);
+    }
+
     #[test]
     fn test_write_enumerated() {
         assert_writes::<Enumerated>(&[
@@ -352,6 +398,11 @@ mod tests {
                 b"\x31\x09\x02\x01\x01\x02\x01\x02\x02\x01\x03",
             ),
         ]);
+
+        assert_writes(&[(
+            parse_single::<SetOf<u64>>(b"\x31\x06\x02\x01\x05\x02\x01\x07").unwrap(),
+            b"\x31\x06\x02\x01\x05\x02\x01\x07",
+        )]);
     }
 
     #[test]
